#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <stdarg.h>
#include <assert.h>
#include <dirent.h>

#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/types.h>

#include <netinet/in.h>

#include <xenctrl.h>

#include "xennerctrl.h"
#include "evtchnd.h"

/* ------------------------------------------------------------------ */

static int wait_data(int fd, int waitsecs)
{
    struct timeval tv, start;
    fd_set rd;
    int rc;

    gettimeofday(&start,NULL);
    for (;;) {
	FD_ZERO(&rd);
	FD_SET(fd,&rd);
	tv.tv_sec  = waitsecs;
	tv.tv_usec = 0;
	rc = select(fd+1, &rd, NULL, NULL, &tv);
	if (rc >= 0)
	    break;
	if (errno != EINTR)
	    break;
	/* got signal, restart? */
	gettimeofday(&tv,NULL);
	if (tv.tv_sec - start.tv_sec > waitsecs) {
	    rc = 0;  /* tiemout */
	    break;
	}
    }
    return rc;
}

static int evtchn_ioctl(int evtchnd, int32_t request, void *data, int no_reply)
{
    struct evtchn_ioctl_msg req;
    struct evtchn_ioctl_msg rsp;
    int size = _IOC_SIZE(request);
    int rc, count = 0;

    if (size > sizeof(req.data)) {
	fprintf(stderr, "%s: data too big: %d/%zd\n",
		__FUNCTION__, size, sizeof(req.data));
	return -1;
    }

    memset(&req, 0, sizeof(req));
    memcpy(req.data, data, size);
    req.ioctl = request;
    rc = write(evtchnd, &req, sizeof(req));
    if (rc != sizeof(req)) {
	fprintf(stderr, "%s: write: %d/%zd\n", __FUNCTION__, rc, sizeof(req));
	return -1;
    }
    if (no_reply)
	return 0;

retry:
    if (!wait_data(evtchnd, 5)) {
	fprintf(stderr, "%s: no reply within 5 sec\n", __FUNCTION__);
	return -1;
    }
    rc = read(evtchnd, &rsp, sizeof(rsp));
    if (rc != sizeof(rsp)) {
	if (-1 == rc && (EAGAIN == errno || EINTR == errno) && count++ < 5) {
	    fprintf(stderr, "%s: read: retrying ... (%s)\n",
		    __FUNCTION__, strerror(errno));
	    goto retry;
	}
	fprintf(stderr, "%s: read: %d/%zd (%s)\n", __FUNCTION__,
		rc, sizeof(rsp),
		(-1 == rc) ? strerror(errno) : "partial read");
	return -1;
    }
    if (req.ioctl != rsp.ioctl) {
	if (rsp.ioctl == EVTCHND_NOTIFY) {
	    if (libxc_trace)
		fprintf(stderr, "%s: got async notify, push back\n", __FUNCTION__);
	    write(evtchnd, &rsp, sizeof(rsp));
	} else if (libxc_fixme) {
	    fprintf(stderr, "%s: FIXME: unexpected reply\n", __FUNCTION__);
	}
	goto retry;
    }

    memcpy(data, rsp.data, size);
    if (-1 == rsp.retval)
	errno = rsp.error;
    return rsp.retval;
}

int xc_evtchn_open(void)
{
    struct sockaddr_in in;
    int rc, evtchnd;

    if (-1 == (evtchnd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP))) {
	perror("socket");
	rc = -1;
	goto out;
    }
    in.sin_family      = AF_INET;
    in.sin_port        = htons(EVTCHND_PORT);
    in.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    if (-1 == connect(evtchnd, (struct sockaddr*) &in, sizeof(in))) {
	perror("connect");
	close(evtchnd);
	rc = -1;
	goto out;
    }
    rc = evtchnd;

out:
    if (libxc_trace)
	fprintf(stderr, "libxc: %s -> handle %d\n",
		__FUNCTION__, rc);
    return rc;
}

int xc_evtchn_close(int evtchnd)
{
    close(evtchnd);
    return 0;
}

int xc_evtchn_fd(int evtchnd)
{
    return evtchnd;
}

evtchn_port_t xc_evtchn_bind_unbound_port(int evtchnd, int domid)
{
    struct ioctl_evtchn_bind_unbound_port io = {
	.remote_domain = domid,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_BIND_UNBOUND_PORT, &io, 0);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, domid %d) -> %d\n",
		__FUNCTION__, evtchnd, domid, rc);
    return rc;
}

evtchn_port_t xc_evtchn_bind_interdomain(int evtchnd, int domid,
					 evtchn_port_t remote_port)
{
    struct ioctl_evtchn_bind_interdomain io = {
	.remote_domain = domid,
	.remote_port   = remote_port,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_BIND_INTERDOMAIN, &io, 0);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, domid %d, rport %d) -> %d\n",
		__FUNCTION__, evtchnd, domid, remote_port, rc);
    return rc;
}

evtchn_port_t xc_evtchn_bind_virq(int evtchnd, unsigned int virq)
{
    struct ioctl_evtchn_bind_virq io = {
	.virq = virq,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_BIND_VIRQ, &io, 0);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, virq %d) -> %d\n",
		__FUNCTION__, evtchnd, virq, rc);
    return rc;
}

int xc_evtchn_unbind(int evtchnd, evtchn_port_t port)
{
    struct ioctl_evtchn_unbind io = {
	.port = port,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_UNBIND, &io, 0);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, port %d) -> %d\n",
		__FUNCTION__, evtchnd, port, rc);
    return rc;
}

int xc_evtchn_notify(int evtchnd, evtchn_port_t port)
{
    struct ioctl_evtchn_notify io = {
	.port = port,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHN_NOTIFY, &io, 1);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, port %d) -> %d\n",
		__FUNCTION__, evtchnd, port, rc);
    return rc;
}

evtchn_port_t xc_evtchn_pending(int evtchnd)
{
    struct evtchn_ioctl_msg msg;
    struct evtchnd_port *n = (void*)(&msg.data);
    int rc, count = 0;

again:
    rc = read(evtchnd, &msg, sizeof(msg));
    if (rc != sizeof(msg)) {
	if (-1 == rc && (EAGAIN == errno || EINTR == errno) && count++ < 5)
	    goto again;
	fprintf(stderr, "libxc: %s: read error (%d/%zd) %s\n",
		__FUNCTION__, rc, sizeof(msg), strerror(errno));
	return -1;
    }
    if (EVTCHND_NOTIFY != msg.ioctl) {
	fprintf(stderr, "libxc: %s: Huh? msg != notify?\n", __FUNCTION__);
	return -1;
    }
    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d) -> %d\n",
		__FUNCTION__, evtchnd, n->port);
    return n->port;
}

int xc_evtchn_unmask(int evtchnd, evtchn_port_t port)
{
    struct evtchn_ioctl_msg msg;
    struct evtchnd_port *n = (void*)(&msg.data);

    memset(&msg, 0, sizeof(msg));
    n->port = port;
    msg.ioctl = EVTCHND_UNMASK;
    write(evtchnd, &msg, sizeof(msg));

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, port %d)\n",
		__FUNCTION__, evtchnd, port);
    return 0;
}


/* ----------------------------------------------------------- */

int xenner_evtchnd_domid(int evtchnd, int domid)
{
    struct ioctl_evtchnd_domid io = {
	.domid = domid,
    };
    int rc = evtchn_ioctl(evtchnd, IOCTL_EVTCHND_DOMID, &io, 0);

    if (libxc_trace)
	fprintf(stderr, "libxc: %s(handle %d, domid %d) -> %d\n",
		__FUNCTION__, evtchnd, domid, rc);
    return rc;
}

